/*
 * $Id: HideableCustomizable.java,v 1.10 2006/03/17 18:14:35 spindler Exp $
 *
 * Copyright (c) 2000-2003 by Rodney Kinney
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License (LGPL) as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, copies are available
 * at http://www.opensource.org.
 */
package downtown;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Composite;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.AffineTransform;
import java.io.File;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

import VASSAL.Info;
import VASSAL.build.GameModule;
import VASSAL.build.module.PlayerRoster;
import VASSAL.build.module.PlayerRoster.Entry;
import VASSAL.build.module.documentation.HelpFile;
import VASSAL.command.ChangeTracker;
import VASSAL.command.Command;
import VASSAL.configure.ColorConfigurer;
import VASSAL.configure.HotKeyConfigurer;
import VASSAL.counters.Decorator;
import VASSAL.counters.EditablePiece;
import VASSAL.counters.GamePiece;
import VASSAL.counters.KeyCommand;
import VASSAL.counters.PieceEditor;
import VASSAL.counters.Properties;
import VASSAL.tools.SequenceEncoder;

/**
 * Add invisibility to a game piece. Based on a chosen strategy the game piece will become invisible
 * for some players. The class utilizes the strategy pattern with hideAccess being the strategy and
 * implementations of HideAccessStrategy a concrete strategy.
 * TODO: Implement piece editor in a more OO way.
 */
public class HideableCustomizable extends Decorator implements EditablePiece {

	private static boolean allHidden;

	public static final String ID = "hide;";

	public static final String HIDDEN_BY = "hiddenBy";

	protected String hiddenBy;

	protected KeyStroke hideKey;

	protected String command = "Invisible";

	private HideAccessInterface hideAccess = null;

	protected Color bgColor;

	protected KeyCommand[] commands;

	protected KeyCommand hideCommand;

	public void setProperty(Object key, Object val) {
		if (HIDDEN_BY.equals(key)) {
			hiddenBy = (String) val;
		} else {
			super.setProperty(key, val);
		}
	}

	public Object getProperty(Object key) {
		if (HIDDEN_BY.equals(key)) {
			return hiddenBy;
		} else if (Properties.INVISIBLE_TO_ME.equals(key)) {
			return invisibleToMe() ? Boolean.TRUE : Boolean.FALSE;
		} else if (Properties.INVISIBLE_TO_OTHERS.equals(key)) {
			return invisibleToOthers() ? Boolean.TRUE : Boolean.FALSE;
		} else {
			return super.getProperty(key);
		}
	}

	public HideableCustomizable() {
		this(ID + "I", null);
	}

	public HideableCustomizable(String type, GamePiece p) {
		setInner(p);
		mySetType(type);
	}

	public void mySetType(String type) {
		SequenceEncoder.Decoder st = new SequenceEncoder.Decoder(type, ';');
		st.nextToken();
		hideKey = st.nextKeyStroke('I');
		command = st.nextToken("Invisible");
		bgColor = st.nextColor(null);
		hideAccess = new HideAccess(st.nextToken("Player")); //, st.nextToken(""));
		commands = null;
		
	}

	public void mySetState(String in) {
		hiddenBy = "null".equals(in) ? null : in;
	}

	public String myGetType() {
		SequenceEncoder se = new SequenceEncoder(';');
		se.append(hideKey).append(command).append(bgColor).append(
				hideAccess.myGetType());
		return ID + se.getValue();
	}

	public String getHideAccessType() {
		return hideAccess.myGetType();
	}

	public String myGetState() {
		return hiddenBy == null ? "null" : hiddenBy;
	}

	public boolean invisibleToMe() {
		return hiddenBy != null
				&& (allHidden || !hideAccess.hasAccess(hiddenBy));
	}

	public boolean invisibleToOthers() {
		return hiddenBy != null
				&& (allHidden || hideAccess.hasAccess(hiddenBy));
	}

	public Shape getShape() {
		if (invisibleToMe()) {
			return new Rectangle();
		} else {
			return piece.getShape();
		}
	}

	public Rectangle boundingBox() {
		if (invisibleToMe()) {
			return new Rectangle();
		} else {
			return piece.boundingBox();
		}
	}

	public void draw(Graphics g, int x, int y, Component obs, double zoom) {
		if (invisibleToMe()) {
			return;
		} else if (invisibleToOthers()) {
			if (bgColor != null) {
				if (Info.is2dEnabled()) {
					Graphics2D g2d = (Graphics2D) g;
					g.setColor(bgColor);
					AffineTransform t = AffineTransform.getScaleInstance(zoom,
							zoom);
					t.translate(x / zoom, y / zoom);
					g2d.fill(t.createTransformedShape(piece.getShape()));
				} else {
					Rectangle r = (Rectangle) piece.getShape();
					g.setColor(bgColor);
					g.fillRect(x + (int) (zoom * r.x), y + (int) (zoom * r.y),
							(int) (zoom * r.width), (int) (zoom * r.height));
				}
			}
			if (g instanceof Graphics2D) {
				Graphics2D g2d = (Graphics2D) g;
				Composite oldComposite = g2d.getComposite();
				g2d.setComposite(AlphaComposite.getInstance(
						AlphaComposite.SRC_OVER, 0.3F));
				piece.draw(g, x, y, obs, zoom);
				g2d.setComposite(oldComposite);
			}
		} else {
			piece.draw(g, x, y, obs, zoom);
		}
	}

	public String getName() {
		if (invisibleToMe()) {
			return "";
		} else if (invisibleToOthers()) {
			return piece.getName() + "(" + command + ")";
		} else {
			return piece.getName();
		}
	}

	public KeyCommand[] myGetKeyCommands() {
		if (commands == null) {
			hideCommand = new KeyCommand(command, hideKey, Decorator
					.getOutermost(this));
			if (command.length() > 0 && hideKey != null) {
				commands = new KeyCommand[] { hideCommand };
			} else {
				commands = new KeyCommand[0];
			}
		}
		return commands;
	}

	public Command myKeyEvent(KeyStroke stroke) {
		myGetKeyCommands();
		if (hideCommand.matches(stroke)) {
			ChangeTracker tracker = new ChangeTracker(this);
			if (invisibleToOthers()) {
				hiddenBy = null;
			} else if (!invisibleToMe()) {
				// hiddenBy = GameModule.getUserId();
				hiddenBy = hideAccess.getSecret();
			}
			return tracker.getChangeCommand();
		}
		return null;
	}

	public String getDescription() {
		// return "Invisible";
		return "Invisible (" + hideAccess.myGetType() + ")";
	}

	public HelpFile getHelpFile() {
		File dir = VASSAL.build.module.Documentation.getDocumentationBaseDir();
		dir = new File(dir, "ReferenceManual");
		try {
			return new HelpFile(null, new File(dir, "Hideable.htm"));
		} catch (MalformedURLException ex) {
			return null;
		}
	}

	public PieceEditor getEditor() {
		return new Ed(this);
	}

	/**
	 * If true, then all hidden pieces are considered invisible to all players.
	 * Used to temporarily draw pieces as they appear to other players
	 * 
	 * @param allHidden
	 */
	public static void setAllHidden(boolean allHidden) {
		HideableCustomizable.allHidden = allHidden;
	}

	private static class Ed implements PieceEditor {
		private HotKeyConfigurer hideKeyInput;

		private JTextField hideCommandInput;
		private JTextField hideStrategyOptionsInput;
//		private JComboBox hideOptionsInput;
		private JComboBox hideStrategyChoice;

		private ColorConfigurer colorConfig;

		private JPanel controls;

		public Ed(HideableCustomizable p) {
			controls = new JPanel();
			controls.setLayout(new BoxLayout(controls, BoxLayout.Y_AXIS));

			hideKeyInput = new HotKeyConfigurer(null, "Keyboard command:  ",
					p.hideKey);
			controls.add(hideKeyInput.getControls());

			Box b = Box.createHorizontalBox();
			hideCommandInput = new JTextField(16);
			hideCommandInput
					.setMaximumSize(hideCommandInput.getPreferredSize());
			hideCommandInput.setText(p.command);
			b.add(new JLabel("Menu Text:  "));
			b.add(hideCommandInput);
			controls.add(b);

			b = Box.createHorizontalBox();
			b.add(new JLabel("Mode: "));
			
			String strategies[] = HideAccess.getStrategies();

			hideStrategyChoice = new JComboBox(strategies);
			
			for (int i = 0; i < strategies.length; i++) {
				if (p.getHideAccessType().startsWith(strategies[i])) {
					hideStrategyChoice.setSelectedIndex(i);
				}
			}			
			hideStrategyChoice.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent evt) {
					if ("comboBoxChanged".equals(evt.getActionCommand())) {
						hideStrategyOptionsInput.setEnabled(HideAccess.hasOption(hideStrategyChoice.getSelectedIndex()));
					}
				}
				
			});

			b.add(hideStrategyChoice);
 		    controls.add(b);
			
			// Simple solution to provide additional information for the strategy chosen
			// Receive additional information from strategy type
			SequenceEncoder.Decoder st = new SequenceEncoder.Decoder(p.getHideAccessType(), ':');
			// Skip strategy
			st.nextToken();
			// Extract information if any
			String information = st.nextToken("Any options");

 		    hideStrategyOptionsInput = new JTextField(16);
			hideStrategyOptionsInput.setMaximumSize(hideStrategyOptionsInput
					.getPreferredSize());
			hideStrategyOptionsInput.setText(information);

//			hideOptionsInput = new JComboBox(getSides());
			
			b = Box.createHorizontalBox();
			b.add(new JLabel("Options: "));
			b.add(hideStrategyOptionsInput);

			b.setEnabled(p.hideAccess.hasOption());
			controls.add(b);

			colorConfig = new ColorConfigurer(null, "Background color",
					p.bgColor);
			controls.add(colorConfig.getControls());
		}


		public String getState() {
			return "null";
		}

		public String getType() {
			SequenceEncoder se = new SequenceEncoder(';');
			se.append((KeyStroke) hideKeyInput.getValue())
			  .append(hideCommandInput.getText())
			  .append(colorConfig.getValue() == null ? "" : colorConfig.getValueString())
			  .append(getHideStrategyType());
			return ID + se.getValue();
		}

		public Component getControls() {
			return controls;
		}

		public String getHideStrategyType() {
			String strategy = HideAccess.getStrategies()[hideStrategyChoice.getSelectedIndex()]
			                                             .concat(":")
						                                 .concat(hideStrategyOptionsInput.getText());
			return strategy;
		}
	}
}



